unit CourbAff;                                 {30/07/2011}
//            Ecran sur lequel les courbes sont traces
interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ExtCtrls, StdCtrls, Math;

type
  Txy = Double;
  TFoncX = Function(X : double) : double of object; // fonction Y = f(X)
  EDomainEr = class(Exception);

  TFAffGraph = class(TForm)
    PBGraph: TPaintBox;
    LaffGr: TLabel;
    LaTitre: TLabel;
    procedure PBGraphPaint(Sender: TObject);
    procedure FormResize(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Dclarations prives}
    UnX, unY : Txy;  //real;     // valeur de l'unit (X ou Y) en pixels
    Xn, Xm, Yn, Ym : Txy;        // valeurs min, max de X et Y et abcisse/ordonne
    Xorig : Txy;                 // valeurs de l'origine des axes
    PXOrig, PYOrig : Integer;    // coordonnes de l'origine en pixels
    A, B : longint;              // largeur et hauteur de la fentre d'affichage
    Xmin, Xmax : Txy;    //modif
    H : longint;         // dcalage vers le bas

    X, Y : Txy;            //abscice et ordon. de la fonction
    Pt : TPoint;           // point de l'cran
    INumAff : integer;     // numro de l'affichage des graduations
//    BOrtho : boolean;    // vrai si repre orthonorme (sur axes unit X=unit y)

    procedure GrAxes(Fx :TFoncX);                        // graduation des axes
    procedure TraceAxes(X0,Y0 : Integer);            // trac des axes en pixels
    procedure TraceGraphXY( Fx : TFoncX; ChY : string; Nb : integer); // trac de Y= f(X)
    procedure TraceDonneesReg( NbSerie : integer); // trace une croix au point Pt
    procedure ParamFonctionX(Xnd, Xmd : Txy; Borthod : boolean);
                                       // calcul des paramtres  units, origine
    procedure PtXY(U, V : Txy);   // coordonnes graphiques d'un point en pixels
    function MakeStr(const Args: array of const): string;
    function ChYfX(StlCo : TStringList): Ansistring;
    function CalYfX( X : Double): Double;        // calcul de Y en fonction de X
  public
    { Dclarations publiques}
    BFDonPoly : Boolean;   // Vrai si appel par FCalculPoly et Faux si FGraphPoly
       // c'est un cas particulier d'utilisation : TFGraphPoly = class(TFDonPoly)

    Xdtab, Ydtab : array of Txy;         // valeurs des donnes  (X, Y) initiales
    procedure AfficheGraphe;                       // prparation de l'affichage
  end;

var
  FAffGraph: TFAffGraph;

implementation

{$R *.DFM}
uses   RegresPoly, RegGraphePoly;

const  M = 15;                // marges pour l'affichage

function TFAffGraph.CalYfX( X : Double): Double;  // calcul de Y = B0 + Sum(Bi * X^i) pour X
var I : integer;                              // les coef. B0, Bi sont dans la liste StlCoef
    Val : Double;
begin
if BFDonPoly then
  with FCalculPoly do begin
    Val :=  StrToFloat(StlCoef[0]);
    for I := 1 to StlCoef.Count -1 do
                         Val := Val + StrToFLoat(StlCoef[I]) * IntPower(X,I);
    Result := Val;
  end
  else with FGraphPoly do begin
    Val :=  StrToFloat(StlCoef[0]);
    for I := 1 to StlCoef.Count -1 do
                         Val := Val + StrToFLoat(StlCoef[I]) * IntPower(X,I);
    Result := Val;
  end;
end;

function  TFAffGraph.ChYfX(StlCo : TStringList): Ansistring;
var
  I : integer;
  ChY : Ansistring;
begin
  ChY := FloatToStrF(StrToFloat(StlCo[0]),ffFixed,5,4); //formatage du nombre
  for I := 1 to StlCo.Count-1 do if StrToFLoat(StlCo[I]) <> 0 then
  begin
    if I=1 then
    ChY := ChY +' + (' + FloatToStrF(StrToFLoat(StlCo[I]), ffFixed,5,4) + ') * X'
    else
    ChY := ChY +' + (' + FloatToStrF(StrToFLoat(StlCo[I]),ffFixed, 5,4) + ')* X^'
                                                                 + IntToStr(I);
  end;
  Result := ChY;
end;

procedure TFAffGraph.PBGraphPaint(Sender: TObject);         // lance l'affichage
var ChYx : string ;
begin
   with Sender as TPaintBox do
     if BFDonPoly  then  begin        //le graphe est demand au calcul initial
        ChYx := ChYfX(FcalculPoly.StlCoef);
        TraceGraphXY(CalYfX, ChYx,FcalculPoly.NbSeries );
      end
      else begin                   // le graphe est demand seul  part
        ChYx := ChYfX(FgraphPoly.StlCoef);
        TraceGraphXY(CalYfX, Chyx, FGraphPoly.NbSeries);
      end;
end;

procedure TFAffGraph.FormResize(Sender: TObject);
begin
   PBGraph.Width  := ClientWidth - 20;
   PBGraph.Height := ClientHeight - PBGraph.Top-10;

   A := PBGraph.Width ;
   B := PBGraph.Height;
   UnX := ((A-2*M) /(Xmax-Xmin));
   UnY := ( (B-2*M-H) /(Ym - Yn) );
   PXorig := Trunc((Xorig-Xmin)*Unx)+M;
   PYorig := Round(Ym*unY) + M + H;
end;

procedure TFAffGraph.FormClose(Sender: TObject; var Action: TCloseAction);
begin
  Close;
end;

procedure TFAffGraph.PtXY(U, V : Txy);  // coordonnes graphiques d'un point en pixels
begin
   Pt.X := Round(PXorig + (U-Xorig) * unX);
   Pt.Y := Round(PYorig - V * unY);
end;

//==================================================================================
procedure TFAffGraph.GrAxes(Fx :TFoncX);   // graduation des axes - c'est empirique
var
  Dx, Dy : Single;         // unit pour la graduation des axes

procedure GraduPtX(U,V : Txy);    // d'un point sur X
var D : integer;           //dcalage horizontal de l'affichage
    Ch : string;           // nombre affich
begin
  with PBGraph.Canvas do
  begin
  PtXY(U, V);                    // cordonnes du point d'affichage  Pt.X Pt.Y
  if Round(U)= U then Ch := FloatToStr(U)
                 else begin
                        Str(U:7:2, Ch);
                        if U<10 then Str(U:5:2, Ch);
                      end;
  if (V=Yn) or (Yn = 0) or ( (U<>0) and(Fx(U)*UnY < 0))
        then D := -20
        else D := 5;           //10
  if (U=Xn) and (U=Xorig) then D := -20;
  MoveTo(Pt.X,Pt.Y+3);
  LineTo(Pt.X,Pt.Y-6);
  if (U=Xm) and ((A-Pt.X-2*M)<TextWidth(Ch)) then Pt.X := Pt.X - TextWidth(Ch);
  if U<>0 then TextOut(Pt.X,Pt.Y+D, Ch);
  end;
end;

procedure GraduPtY(U,V : Txy);  // d'un point sur Y
var D : integer;
    Ch : string;
begin
  with PBGraph.Canvas do
  begin
    PtXY(U, V);               // coordonnes en pixels
    MoveTo(Pt.X+3,Pt.Y);
    LineTo(Pt.X+3,Pt.Y);
    LineTo(Pt.X-6,Pt.Y);
    if Pt.Y >= (B-M-2) then D := -10
                       else D := -2;
    if Abs(V) < 1.e-06 then V := 0;
    if V = Ym then D := 0;
    Ch := FloatToStrF(V,ffFixed,7,3);
    if (V=0) and (U=XOrig) then           // graduation de l'intersection des axes
                              TextOut(Pt.x+3,Pt.y+1, '('+ FloatToStr(U) + ', 0.0)')
                            else
                             TextOut(Pt.x+8,Pt.y+D, Ch );
  end;
end;

function Unite(Un, Um : Txy): Single;// calcul de l'unit optimale pour les graduations
var U : Txy;
    N : integer;
begin
  U := Log10(Abs(Um - Un)/4);
  if U < 0 then U := U - 1;
  N := Trunc(U);
  Result := Power(10, N);
end;

procedure AxeXGr(U, dd : single);  // axe des X
var Cote : integer;
    D : integer;                  //dplacement pour affichage hors axe
begin
  D := 2;
  Cote := round(U/dd);
  if Cote <> 0 then
    if (Cote mod 10)=0 then D := 6
    else if (Cote mod 5)=0 then D := 4 ;
  with PBGraph.Canvas do    // affichage repre'|' sur axe X
  begin
    Ptxy(U, 0);
    MoveTo(Pt.X, Pt.Y-D);
    LineTo(Pt.X, Pt.Y+D);
  end;
  if not Odd(InumAff) then  GraduPtX(U, 0);
  Inc(INumAff);
end;

procedure AxeYGr(U, dd : single);   // axe ds Y
var Cote : integer;
    D : integer;
begin
  D := 2;
  Cote := round(U/dd);
  if Cote <> 0 then
    if (Cote mod 10)=0 then
      begin D := 5;             // le trait sera plus ou moins long ?
          GraduptY(Xorig, U);
      end
    else if (Cote mod 5)=0 then
      begin D := 3;
          GraduptY(Xorig, U);
      end;
  with PBGraph.Canvas do      // marque sur axe Y
  begin
    Ptxy(Xorig, U);
    MoveTo(Pt.X-D, Pt.Y);
    LineTo(Pt.X+D, Pt.Y);
  end;
end;

procedure GraduAxeX(Un, Um : Txy; dd : single);
var U : Txy;                 // unit
begin
  INumAff := 1;
  if Un < Xorig then
  begin
    U := XOrig;
    repeat
       U := U-dd;
       if (U<=Um) then AxeXGr(U, dd);
    until U < Un
  end;
  if Um > XOrig then
  begin
    U := XOrig;
    repeat
      U := U+dd;
      if (U>=Un) then AxeXGr(U, dd);
    until U>Um;
  end;
end;

procedure GraduAxeY(Un, Um : Txy; dd : single);
var U : Txy;
begin
  if Un < 0 then
  begin
    U := 0;
    repeat
       U := U-dd;
       if (U<=Um) then AxeYGr(U, dd);
    until U < Un
  end;
  if Um > 0 then
  begin
    U := 0;
    repeat
      U := U+dd;
      if (U>=Un) then AxeYGr(U, dd);
    until U>Um;
  end;
  GraduPtY(Xorig, 0);
end;
 //-------------------------------------  GrAxes  ------------------------
begin
  Dx := Unite(Xn, Xm);                     // graduation de l'axe des X
  while (Dx*UnX) < 10 do Dx := Dx * 2;
  GraduAxeX(Xn,Xm,Dx);
  if (Xorig<>0) and (abs(Xorig)>4) then GraduPtX(0.0,0);
  Dy := Unite(Yn, Ym);                    // graduation de l'axe des Y
  GraduAxeY(Yn, Ym, Dy);
end;
//================================================================================


procedure TFAffGraph.TraceAxes(X0,Y0 : Integer);   // trac des axes en pixels
var W,H : Integer;
begin
   W := PBGraph.Width;
   H := PBGraph.Height;
   PBGraph.Canvas.MoveTo(0,Y0);   // axe x
   PBGraph.Canvas.LineTo(W,Y0);
   PBGraph.Canvas.MoveTo(X0,0);   // axe y
   PBGraph.Canvas.LineTo(X0,H-2);
end;

           //trac de Y= f(X)  entre Xn..Xm
procedure TFAffGraph.TraceGraphXY( Fx : TFoncX; ChY : string; Nb : integer);
var                            //             chaine Y=F(X)  Nbre Series de point
  Pas : Txy;
  BPtdsEcran : boolean;       // faux si 1er point
begin
//  with AffGraph do
  begin
    BPtdsEcran := false;
    Pas := (Xm-Xn)/2/A;
    X := Xn;
    while X <= Xm do
    begin
      try
        Y := Fx(X);
      except
        on E: EInvalidOp do
        begin
          MessageDlg(MakeStr(['CourbAff.FonctionXY -Valeur non dfinie pour X = ',X,
                      #13#10,E.Message ,'   ',E.HelpContext]), mtWarning,[mbOK],0);
          Exit;
        end;
      end;
      PtXY(X, Y);
      if (Pt.Y >=0) and (Pt.Y<= B) then
        if BPtdsEcran then PBGraph.Canvas.LineTo(Pt.X, Pt.Y)
                      else begin
                             PBGraph.Canvas.MoveTo(Pt.X, Pt.y);
                             BPtdsEcran := true;
                      end
        else BPtdsEcran := false;
      X := X + Pas;
    end;
    LaTitre.Caption := 'Reg. Polynomiale - '+FCalculPoly.EdTitre.Text+
             ' ( Fichier : '+ FCalculPoly.EdNomFic.Text+')';
    LaffGr.Caption := 'Fonction f(x) = '+ ChY
           +' - Intervalle ('+FloatToStr(Xmin)+' .. '+FloatToStr(Xmax)+')';

    if Nb > 0 then TraceDonneesReg( Nb);               // affiche les points donnes
    TraceAxes(PXOrig,PYOrig);
    GrAxes(Fx);
  end;
end;

procedure TFAffGraph.ParamFonctionX(Xnd, Xmd : Txy; Borthod : boolean);
                    //   X min, max      si Vrai repre orthonorm (unites axes X=Y)
                  // calcul des paramtres  units, origine
var W : Txy;
begin
  if Xnd >= Xmd then begin         // correction si Xnd n'est pas infrieur  Xmd
                     W := Xnd;
                     Xnd := Xmd;
                     Xmd := W;
                   end;
  Xmin := Xnd;    Xmax := Xmd;
  if (Xmin<0) and (Xmax>0) then Xorig := 0
                           else Xorig := Xmin;
  H := 0;
  if Yn > 0 then Yn := 0;
  if Ym < 0 then begin Ym := 0;
                       H := 10;   end;
  UnX := ( (A-2*M) /(Xmax-Xmin));         // units en points
  UnY := ( (B-2*M-H) /(Ym - Yn) );
  PXorig:= Trunc((Xorig-Xmin)*Unx)+M;     // position du point origine (Xorig,...)
  PYorig := Round(Ym*unY) + M + H;        // sur l'cran - H dcalage vers le haut
  if BOrthod then                         // si vrai mme unit sur les axes
    if UnX < UnY then  UnY := UnX
                 else  UnX := UnY;
  ShowModal;                     // excute PBGraphPaint
end;

procedure TFAffGraph.AfficheGraphe;      // prparation de l'affichage
begin                         // calcul des limites de l'affichage
  Xmin := MinValue(Xdtab);
  Xmax := MaxValue(Xdtab);
  Xmin := Int(0.9 * Xmin);     // valeur entire min
  Xmax := Int(1.1 *Xmax + 1);  //                max
  if Xmin > 800 then
  begin                       // arrondi des nombres suprieurs  800  
    Xmin := Int(Xmin/10)*10;
    Xmax := Int(Xmax/10+1)*10;
  end;
  Yn := Minvalue(Ydtab);
  Ym := Maxvalue(Ydtab);
  Xn := Xmin;
  Xm := Xmax;
  A := PBGraph.Width;
  B := PBGraph.Height;
  ParamFonctionX( Xmin, Xmax, False);   // units et origine
end;

procedure TFAffGraph.TraceDonneesReg( NbSerie : integer);   // trace une croix au point Pt
var I : integer ;             //   Nbre de sries de donnes dfinissant un point (X,Y)

procedure TraceCroix;   // trace un  +
var D : integer;
begin
  with PBGraph.Canvas do
  begin
    D := 4;
    MoveTo(Pt.X-D, Pt.Y);
    LineTo(Pt.X + D, Pt.Y);
    MoveTo(Pt.X, Pt.Y-D);
    LineTo(Pt.X, Pt.Y + D);
  end;
end;

begin
  for I := 0 to NbSerie-1 do   
  begin
//    with AffGraph do
    begin
      PtXY(Xdtab[I], Ydtab[I]);
      PBGraph.Canvas.MoveTo(Pt.X, Pt.Y);
      TraceCroix;
    end;
  end;
end;     //------------- TraceDonneesReg--------------------------------

function TFAffGraph.MakeStr(const Args: array of const): string;  // Delphi 3 
// Exemple, en partant de la fonction MakeStr ci-dessus, l'instruction
//    S := MakeStr(['Test ', 100, '-', True, '-', 3.14159]);
//   gnre la chane suivante:  'Test 100-T-3.14159'
const
 BoolChars: array[Boolean] of Char = ('F', 'T');
var
 I: Integer;
begin
 Result := '';
 for I := 0 to High(Args) do
   with Args[I] do
     case VType of
       vtInteger:    Result := Result + IntToStr(VInteger);
       vtBoolean:    Result := Result + BoolChars[VBoolean];
       vtChar:       Result := Result + VChar;
       vtExtended:   Result := Result + FloatToStr(VExtended^);
       vtString:     Result := Result + VString^;
       vtPChar:      Result := Result + VPChar;
       vtObject:     Result := Result + VObject.ClassName;
       vtClass:      Result := Result + VClass.ClassName;
       vtAnsiString: Result := Result + string(VAnsiString);
       vtCurrency:   Result := Result + CurrToStr(VCurrency^);
       vtVariant:    Result := Result + string(VVariant^);
     end;
end;

end.



